/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include <iostream>
#include <cstdlib>

#include <xercesc/util/PlatformUtils.hpp>

#include "wordlist.hpp"
#include "bufferedsettings.hpp"
#include "wordtag.hpp"

#include "sounderror.hpp"
#include "inputreader.hpp"
#include "errorintroducer.hpp"
#include "splitcompound.hpp"
#include "damerau.hpp"
#include "driver.hpp"

XERCES_CPP_NAMESPACE_USE

void startupXML() {
    try {        
        XMLPlatformUtils::Initialize();
    }
    catch (const XMLException& toCatch) {
        char* message = XMLString::transcode(toCatch.getMessage());
        std::cout << "Error during initialization! :\n"
                  << message << "\n";
        delete [] message;
        std::exit(1);
    }
}

void shutdownXML() {
    XMLPlatformUtils::Terminate();
}

void printUsage(const char *name) {
    std::cerr << "Usage:  " << name << " [ options... ]\n"
              << "Options:\n"
              << "   -f<filename>\tSpecifies which settings file to use\n"
              << "   -i<filename>\tSpecifies the input filename\n"
              << "   -o<filename>\tSpecifies the output filename\n"
              << "   -x<filename>\tSpecifies the name of the error XML description\n"
              << "   -n<seed>\tSpecifies the seed for the random number generator\n"
              << "   -d<rate>\tEnable the Damerau module\n\t\tSpecifying <rate> overrides default error rate\n"
              << "   -s<rate>\tEnable the SplitCompound module\n\t\tSpecifying <rate> overrides default error rate\n"
              << "   -e<rate>\tEnable the Sound Error module\n\t\tSpecifying <rate> overrides default error rate\n"
              << "   +d\t\tDisable the Damerau module\n"
              << "   +s\t\tDisable the SplitCompound module\n"
              << "   +e\t\tDisable the Sound Error module\n"
              << "   -h\t\tThis help text\n";
    std::exit(0);
}

int main(int argc, char* argv[]) {
    startupXML();
    std::string settingsfile = "settings.xml";

    for (int i = 1; argv[i]; ++i) {
        char *arg = argv[i];
        if (*arg == '-') {
            switch (*++arg) {
            case 'f': 
                if (*++arg == '\0') printUsage(argv[0]); 
                settingsfile = arg; break;
            case 'i':
            case 'o':
            case 'x':
            case 'n':
            case 'd':
            case 's':
            case 'e': 
            case '-': break;
            case 'h':
            default: printUsage(argv[0]); break;
            }
        }
    }
    BufferedSettings settings(settingsfile);
    if (!settings.insertCommandLineParameters(argc, argv)) {        
        printUsage(argv[0]);
    }    
    if (settings.getString("options/configDir", "") != "") {
        settings.addParameter("options/configDir", settings.getString("options/configDir", "../conf") + "/");
    }
    Wordlist wordlist(settings);
    ErrorWriter errorwriter(wordlist, settings);
    errorwriter.insertInfo(argv, settings.getString("options/seed", "123456789"), settingsfile);

    InputReader inputreader(wordlist, settings);

    IntroducerList introducerlist(wordlist, settings, errorwriter, inputreader);
    SoundError sounderror(wordlist, settings, errorwriter);
    SplitCompound splitCompound(wordlist, settings, errorwriter, inputreader);
    Damerau damerau(wordlist, settings, errorwriter);

    Driver driver(wordlist, settings, inputreader, damerau, splitCompound, introducerlist, sounderror);
    driver.createErrors();

    errorwriter.write();
    shutdownXML();
    std::string hej;
    std::cerr << "Goodbye.\n";
    return 0;
}
