/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

//
// This class computes the minimum distance between two strings
//
// distance  - insertion, deletion and substitution
//
// distance2 - insertion and deletion
//
// Based on an minimum edit distance 
// implementation by Anders Sewerin Johansen
// http://www.merriampark.com/ldcpp.htm,
// then heavily modified to fit its purpose here
//

#include <iostream>
#include <list>

#include <boost/tuple/tuple.hpp>

#include "matrix.hpp"

class MinEdit {
private:
     template<class T>
     static T min(T t1, T t2) { return t1 < t2 ? t1 : t2; }

     template<class T>
     static T max(T t1, T t2) { return t1 > t2 ? t1 : t2; }

     template<class T>
     static bool between(T elem, T t1, T t2) {
	  return elem >= min(t1, t2) && elem <= max(t1, t2);
     }

     class Data {
     public:
	  Data() : num(0), above(false), left(false), diag(false) {}
	  int num;
	  bool above;
	  bool left;
	  bool diag;
     };

public:

     typedef std::list<boost::tuple<char, int, int> > OperationList;

     // FIXME byt ordning p src och dst...
     static int distance(const std::string src, const std::string dst) {
	  const int n = src.length();
	  const int m = dst.length();
	  if (n == 0) return m;
	  if (m == 0) return n;
	  
	  Matrix<int> matrix(n + 1, m + 1);
	  
	  for (int i = 0; i <= n; i++) matrix[i][0] = i;
	  for (int j = 0; j <=m; j++) matrix[0][j] = j;

	  for (int i = 1; i <= n; i++) {
	       for (int j = 1; j <= m; j++) {
		    int cost;
		    if (src[i - 1] == dst[j - 1]) {
			 cost = 0;
		    }
		    else {
			 cost = 1;
		    }
		    const int above = matrix[i-1][j];
		    const int left = matrix[i][j-1];
		    const int diag = matrix[i-1][j-1];
		    const int cell = min(above + 1, min(left + 1, diag + cost));
		    matrix[i][j] = cell;
	       }
	  }
	  /*
	  std::cerr << "   " << dst << std::endl << "  ";
	  for (int i = 0; i <=m; i++) {
	       std::cerr << i;
	  }
	  std::cerr << std::endl;
	  for (int j = 1; j <= n; j++) {
	       std::cerr << src[j-1] << " " << j;
	       for (int i = 1; i <= m; i++) {
		    std::cerr << matrix[j][i];
	       }
	       std::cerr << std::endl;
	  }
	  */
	  return matrix[n][m];
     }

     template<class RaContainer>
     static OperationList distance2(const RaContainer src, const RaContainer dst) {
	  const int n = dst.size();
	  const int m = src.size();
	  if (n == 0) {
	       OperationList empty;
	       return empty;
	  }
	  if (m == 0) {
	       OperationList empty;
	       return empty;
	  }
	  
	  Matrix<Data> matrix(n + 1, m + 1);
	  
	  for (int i = 0; i <= n; i++) {
	       matrix[i][0].num = i;
	       matrix[i][0].above = true;
	  }
	  for (int j = 0; j <=m; j++) {
	       matrix[0][j].num = j;
	       matrix[0][j].left = true;
	  }

	  // Fill distance matrix
	  for (int i = 1; i <= n; i++) {
	       for (int j = 1; j <= m; j++) {
		    int cost;
		    if (dst[i - 1] == src[j - 1]) {
			 cost = 0;
		    }
		    else {
			 cost = 2;
		    }
		    const int above = matrix[i-1][j].num;
		    const int left = matrix[i][j-1].num;
		    const int diag = matrix[i-1][j-1].num;
		    Data dataCell;
		    if (above + 1 < left + 1) {			 
			 if (above + 1 < diag + cost) {
			      // above + 1 minst
			      dataCell.num = above + 1;
			      dataCell.above = true;
			 }
			 else {
			      // diag + cost minst
			      dataCell.num = diag + cost;
			      dataCell.diag = true;
			 }
		    }
		    else {
			 if (left + 1 < diag + cost) {
			      // left + 1 minst
			      dataCell.num = left + 1;
			      dataCell.left = true;
			 }
			 else {
			      // diag + cost minst
			      dataCell.num = diag + cost;
			      dataCell.diag = true;
			 }
		    }
		    matrix[i][j] = dataCell;
	       }
	  }

	  typedef typename RaContainer::value_type value_type;
	  typedef boost::tuple<char, value_type, int, int> TraceTuple;
	  std::list<TraceTuple> trace;

	  /*
	  std::cerr << "   " << src << std::endl << "  ";
	  for (int i = 0; i <=m; i++) {
	       std::cerr << i;
	  }
	  std::cerr << std::endl;
	  for (int j = 1; j <= n; j++) {
	       std::cerr << dst[j-1] << " " << j;
	       for (int i = 1; i <= m; i++) {
		    std::cerr << matrix[j][i].num << "[" << matrix[j][i].above << matrix[j][i].left << matrix[j][i].diag << "] ";
	       }
	       std::cerr << std::endl;
	  }
	  */
	  //std::cerr << "End: " << matrix[n][m].num;	  

	  // Going backwards through the matrix, find the path
	  int j = n;
	  int i = m;
	  while (matrix[j][i].num != 0) {
	       if (matrix[j][i].above) {
		    trace.push_back(boost::make_tuple('i', dst[j-1], i+1, j));
		    j--;
		    
	       }
	       else if (matrix[j][i].left) {
		    trace.push_back(boost::make_tuple('d', src[i-1], i, 0));
		    i--;
	       }
	       else if (matrix[j][i].diag) {
		    if (matrix[j][i].num != matrix[j-1][i-1].num) {
			 trace.push_back(boost::make_tuple('d', src[i-1], i, 0));
			 trace.push_back(boost::make_tuple('i', dst[j-1], i, j));
		    }
		    else {
			 //std::cerr << " Going diag...\n";
		    }
		    j--;
		    i--;
	       }
	       else {
		    std::cerr << "Error at j=" << j << ", i=" << i << "\n";
		    break;
	       }
	  }
	  
	  //std::cerr << "Before compact:\n";
	  //printList(trace);

	  // Convert matching 'insert' and 'delete' commands to 'move' commands
	  OperationList oplist = compactList(trace);

	  //std::cerr << "After compact:\n";
	  //printList(trace);
	  
	  return oplist;
     }

     template<class T>
     static void printList(std::list<boost::tuple<char, T, int, int> > &trace) {
	  typedef std::list<boost::tuple<char, T, int, int> > TraceList;
	  typedef typename TraceList::iterator TraceListIterator;
	  for (TraceListIterator iter = trace.begin(); iter != trace.end(); ++iter) {
	       char operation = boost::get<0>(*iter);
	       T element = boost::get<1>(*iter);
	       int pos = boost::get<2>(*iter);
	       int pos2 = boost::get<3>(*iter);
	       std::cerr << operation << element << pos << pos2 << std::endl;
	  }
     }


     /*
      * Convert a delete command with a matching
      * insert command to a single move command.
      */
     template<class T>
     static OperationList compactList(std::list<boost::tuple<char, T, int, int> > &trace) {
	  typedef std::list<boost::tuple<char, T, int, int> > TraceList;
	  typedef typename TraceList::iterator TraceListIterator;

	  // For each item in the command list
	  for (TraceListIterator iter = trace.begin(); iter != trace.end(); ++iter) {

	       char operation = boost::get<0>(*iter);
	       T element = boost::get<1>(*iter);
	       int pos = boost::get<2>(*iter);

	       // For each item in the rest of the list
	       for (TraceListIterator iter2 = iter; iter2 != trace.end(); /*nothing*/) {
		    bool iter2_changed = false;
		    char operation2 = boost::get<0>(*iter2);
		    T element2 = boost::get<1>(*iter2);
		    int pos2 = boost::get<2>(*iter2);

		    // If elements match
		    if (element == element2) {
			 // First 'delete', then 'insert'
			 if (operation == 'd' && operation2 == 'i') {
			      int diff = 0;
			      // Change 'delete' to 'move'
			      boost::get<0>(*iter) = 'm';
			      // Calculate difference
			      for (TraceListIterator iter3 = iter; iter3 != iter2; ++iter3) {
				   int pos3 = boost::get<2>(*iter3);
				   if (pos3 < min(pos, pos2)) {
					char operation3 = boost::get<0>(*iter3);
					if (operation3 == 'd') {
					     --diff;
					}
					else if (operation == 'i') {
					     ++diff;
					}
				   }
			      }
			      // Remove 'insert' command
			      boost::get<3>(*iter) = pos2 + diff;
			      iter2 = trace.erase(iter2);
			      iter2_changed = true;
			      // Update the items in between if necessary
			      for (TraceListIterator iter3 = iter; iter3 != iter2; ++iter3) {
				   if (iter3 != iter) {
					int pos3 = boost::get<2>(*iter3);
					if (between(pos3, pos, pos2)) {
					     boost::get<2>(*iter3) = pos3 + 1;
					}
				   }
			      }
			      break;
			 }
			 // First 'insert', then 'delete'
			 else if (operation == 'i' && operation2 == 'd') {
			      int diff = 0;
			      // Change 'insert' to 'move'
			      boost::get<0>(*iter) = 'm';
			      boost::get<3>(*iter) = pos - 1; // -1 pga att erase grs senare
			      // Calculate difference
			      for (TraceListIterator iter3 = iter; iter3 != iter2; ++iter3) {
				   int pos3 = boost::get<2>(*iter3);
				   if (pos3 <= min(pos, pos2)) {
					char operation3 = boost::get<0>(*iter3);
					if (operation3 == 'd') {
					     --diff;
					}
					else if (operation == 'i') {
					     ++diff;
					}
				   }
			      }
			      // Remove 'delete' command
			      boost::get<2>(*iter) = pos2 + diff;
			      iter2 = trace.erase(iter2);
			      iter2_changed = true;
			      // Update the items in between if necessary
			      for (TraceListIterator iter3 = iter; iter3 != iter2; ++iter3) {
				   if (iter3 != iter) {
					int pos3 = boost::get<2>(*iter3);
					if (between(pos3, pos, pos2)) {
					     boost::get<2>(*iter3) = pos3 - 1;
					}
				   }
			      }
			      break;
			 }
			 else {
			      //std::cerr << " (ignored)\n";
			 }
		    }
		    // If a conversion was done, we already are at the correct
		    // posistion since an item was removed.
		    if (!iter2_changed) {
			 ++iter2;
		    }
	       }
	  }

	  OperationList oplist;
	  for (TraceListIterator iter = trace.begin(); iter != trace.end(); ++iter) {
	       char operation = boost::get<0>(*iter);
	       int pos = boost::get<2>(*iter);
	       int pos2 = boost::get<3>(*iter);
	       oplist.push_back(boost::make_tuple(operation, pos, pos2));
	  }
	  return oplist;
     }
};
